/************************************************************************/
/*                                                                      */
/*   Program Name :  k5_2_octad                                         */
/*                                                                      */
/*     K5_2_octad is a K5 to OCTAD format converter                     */
/*                                                                      */
/*     Enviroment variables                                             */
/*           M5DIR      ---   default directory for Mark5 data          */
/*           M5VEX      ---   default directory for Mark5 vex file      */
/*                                                                      */
/*            Ver 1.00  2014-06-01 start development                    */
/*                                                                      */
/*      Copyright (c) 2014-     T.Kondo All Right Reserved              */
/*                                                                      */
/************************************************************************/
/*

     This program just removes K5 header


     2014-06-01  -- development started  (only 1ch mode support)

   how to run
          k5_2_octad k5_filename [options]
    
     where   vdif_filename -- VDIF data file name
              options (any order)

               -bs nbitsream  -- set Mark5B # of bit-stream (1,2,4,8,16,32)
               -adbit adbit   -- # of AD conversion bit (1,2) (deflt=1)
               -fsampl fmhz   -- channel sampling frequency (MHz) (deflt=4)
			   -1ch m5ch      -- set K5 data to 1ch mode and pick up Mark5B ch (starting at 1)
			   -4ch m5ch1 m5ch2 m5ch3 m5ch4
			                  -- set K5 data to 4 ch mode and pick up Mark5B 4 channels
			   -2bit mode     -- set 2bit AD decode table
			                         1: offset binary   0,1,2,3 (no conversion)
									 2: Mark V          0,2,1,3  (default)
									 3: signed integer  1,0,2,3

               -o k5name     -- K5 file name to be created
                                 (default : see below)
               -d k5dir      -- directory for k5 file out
                                 (default  : mark5 directory)
               -s[offset] soffset
			         +VE : start offset (sec) (default is 0)
					 -VE : absolute value means skip over bytes 
               -k[ashima] [bytesframe] 
                     set header format to Kashima compulsory 
                     bytesframe: +VE bytes/frame set compulsory
  
  Naming rule for K5 file created (MK5 : original mark5 name)
         4ch mode
            MK5.k5a ---- for group#1 in info_file (usually CH 01-04)
            MK5.k5b ---- for group#2 in info_file (usually CH 05-08)
            MK5.k5c ---- for group#3 in info_file (usually CH 09-12)
            MK5.k5d ---- for group#4 in info_file (usually CH 13-16)
         1ch mode
            MK5.k5-NN ---- where NN is channel number 01-16

  Enviroment variables
          M5DIR    ---   default path to Mark5 data
          M5VEX    ---   default path to Mark5 vex file


  Note:  When compiling on VC++, set stack size at least 4000000(4MB).
         otherwise results in runtime error.


*** sample run parameters ****
*/

#define PROGRAM_VERSION "k5_2_octad (Ver 1.00  2014-06-01)"

#include "libvdif.h"
//#include "libk5.h"      // common header for IP-VLBI utilities

//#include "libmk5tok5.h"  // library for Mark5 to K5 conversion

/* default conversion information file name set */
//#define INFONAME_DFLT "m5tok5info.txt"   // information file default name
/* environment variable for Mark5 data default directory set */
#define M5DIR "M5DIR" // the environment variable for Mark5 data default directory
/* environment variable for Mark5 vex file default directory set */
#define M5VEX "M5VEX" // the environment variable for Mark5 vex file


/* default parameters for run mode 2 */
#define K5NUMCH  4
#define M5PARITY 0
#define M5VLBA   0
#define M5FANOUT 1
#define M5ADBIT  1
#define M5SAMPLE 4   // default sampling frequency in MHz
#define M5NTRACK 32

#define DFLT_DATA_FILE "vdif.dat"
#define BUFMAX 25000   // maximum read buffer size

typedef struct k5_2_octadrunp     // run parameters for mark5 to k5 conversion program
{
    int ptype;   // run parameter type  =1  mode1,  =2 mode 2
    int channel;
    int group;
    int soffset;
    int period;
    int reverse;
	int64_t skip_bytes;   // skip bytes (1st skip over # of bytes)
    int scan;    // scan# for get MODE from VEX file (default=1) 
    int odd;    // =1 4byte integer offset mode in case of 64 track
    int monit;   // =1  inofo_file monitor ON,  =0 no monitoring
    char octadname[256];
    char k5name[256];
    char octadname_given[256];
    char outdir[256];
    char info_file[256];
    char vex_file[256];
    char sid[4];
    char dflt_pass[4];
} k5_2_octadrunp_t;

int convK5_2_OCTAD(k5_2_octadrunp_t *runp);

int cmnd_get_num(int argc, char *argv[], char *command, char *param, int *num)
// get argument like -command parameter
// search "-command" in argument, if it is exsted, 0 is returned as
// a function value, and next argument is returned as a parameter (ascii).
// if "-command" is not exsited, -1 is returned as a function value.
// if "-command" is not accompanied by real parameter, NULL character
// is returned as param.
// 
{
    int n;
    char cbuf[80];

    param[0]='\0';

    //search command
    //printf("%s\n",command); // debug
    n=2;
    while(n <=argc) {
        strcpy(cbuf,*(argv+n-1));
        if(strncmp(cbuf,command,strlen(command))==0){  
            n++;
			*num=n;
            if(n > argc) return 0;
            strcpy(param,*(argv+n-1));
			*num++;
            return 0;
        }
        n++;
    }
    return -1;
}

void param_get_runp(int argc, char *argv[],k5_2_octadrunp_t *runp)                       
{
    int n,i;
    char cbuf[256];

	// initialize parameters

    /* common parameters */
    runp->ptype=1;
    runp->octadname[0]='\0';
    runp->k5name[0]='\0';
    runp->octadname_given[0]='\0';
    runp->soffset=0;
    runp->period=0;
    runp->odd=0;

    /* mode 1 parameters */
    runp->channel=0;
    runp->group=0;
    runp->reverse=0;
    runp->info_file[0]='\0';
    runp->outdir[0]='\0';
    runp->vex_file[0]='\0';
    runp->sid[0]='\0';
    runp->scan=1;
    runp->monit=0;

    /* 1st parameter (vdif file name) get */
    if (argc>=2){
       strcpy(runp->k5name,*(argv+1));
	   if (strlen(*(argv+1))==1 && runp->k5name[0]=='0'){  // set default name
          strcpy(runp->k5name,DFLT_DATA_FILE);
	   }
    }
     /* check K5 file name is acompanied by directory or not */
    get_true_path(runp->k5name, M5DIR, "./",cbuf);
    strcpy(runp->k5name,cbuf);

    /* command parameters get */
    if(cmnd_get(argc,argv,"-o",cbuf)==0) strcpy(runp->octadname_given,cbuf);  // octad name
	if(cmnd_get(argc,argv,"-d",cbuf)==0){              // out dir
		strcpy(runp->outdir,cbuf);  
        /* env variable is substituted by real one */
        if((i=substitute_env_path_name(runp->outdir, cbuf))>0){
           strcpy(runp->outdir,cbuf);
        }
	}
    //if(cmnd_get_p(argc,argv,"-s",cbuf)==0) runp->soffset = atoi(cbuf);  // start offset
    if(cmnd_get(argc,argv,"-p",cbuf)==0) runp->period  = atoi(cbuf);  // period
	if(cmnd_get(argc,argv,"-r",cbuf)==0) runp->reverse=1;  // bit reverse mode
	if(cmnd_get(argc,argv,"-monit",cbuf)==0) runp->monit=1;  // monitor mode on

}

void print_usage(void)
{
    printf("%s\n",PROGRAM_VERSION);
    printf("\n");
    printf("    k5_2_octad k5_file [options]\n");
    printf("\n");
    printf("where   k5_file  -- K5 data file name\n");
    printf("      options (any order)\n");
    printf("        -o k5name     -- K5 file name to be created\n");
    printf("                          (default : see below)\n");
    printf("        -d k5dir      -- directory for k5 file out\n");
    printf("                          (default  : mark5 directory)\n");
    printf("        -bs nbitsream -- set Mark5B # of bit-stream (1,2,4,8,16,32)\n");
    printf("        -fsampl fmhz  -- channel sampling frequency (MHz) (deflt=%d)\n",M5SAMPLE);
    printf("        -adbit adbit  -- # of AD conversion bit (deflt=%d)\n",M5ADBIT);
    printf("        -1ch vdifch     -- set K5 data to 1ch mode and \n");
    printf("       	                    pick up Mark5B ch (starting at 1)\n");
    printf("        -4ch ch1 ch2 ch3 ch4\n");
    printf("                      -- set K5 data to 4 ch mode and  \n");
    printf(" 	                       pick up VDIF 4 channels\n");
    printf("        -2bit mode    -- set 2bit AD decode table\n");
    printf("                           1: offset binary   0,1,2,3 (no conversion)\n");
    printf("                           2: Mark V standard 0,2,1,3 (default)\n");
    printf("                           3: signed integer  1,0,2,3\n");
    printf("        -s[offset] soffset\n");
    printf("                        +VE : start offset (sec) (default is 0.0)\n");
    printf("                        -VE : absolute value means skip over bytes \n");
    printf("        -k[ashima] [bytesframe]\n");
    printf("                      -- set header format to Kashima compulsory \n");
	printf("                         bytesframe: +VE is bytes/frame set compulsory\n");
//    printf("        -s strat_sec  -- offset time from data head (sec)\n");
//    printf("                          (default :  0 )\n");
//    printf("        -p period     -- data period to convert (sec)\n");
//    printf("                          (default : all)\n");
//    printf("        -r            -- reverse track order when fanout =2 or more\n");
//    printf("        -v vex_file   -- VEX file name to be read when \"information\n");
//    printf("                              file creation mode\" is set\n");
//    printf("        -sid stat_id  -- station ID when \"information file creation\n");
//    printf("                              mode\" is set\n");
//    printf("        -scan scan#   -- scan# for MODE get when \"information file\n");
//    printf("                              creation mode\" is set (default : 1)\n");
//    printf("        -subp subpass -- set subpass at information file\n");
//    printf("                              creation mode (default is \"A\")\n");
//    printf("        -odd          -- 32bit (4byte) shift in reading Mark5 data\n");
//    printf("                          (this works only for 64 track mode)\n");
//    printf("        -monit        -- info_file monitor ON\n");
//    printf("\n");
    printf("   Naming rule for K5 file created (VDIF : original VDIF name)\n");
    printf("      4ch mode\n");
    printf("        VDIF.k5a ---- for group#1 (usually ch01-04)\n");
    printf("        VDIF.k5b ---- for group#2 (usually ch05-08)\n");
    printf("        VDIF.k5c ---- for group#3 (usually ch09-12)\n");
    printf("        VDIF.k5d ---- for group#4 (usually ch13-16)\n");
    printf("      1ch mode\n");
    printf("        VDIF.k5-NN ---- where NN is channel number 01-16\n");
    printf("   Enviroment variables\n");
    printf("        %s    ---   default path to Mark5 data ( %s )\n",
        M5DIR,getenv(M5DIR));
    printf("        %s    ---   default path to Mark5 vex file ( %s )\n",
        M5VEX,getenv(M5VEX));
    printf("\n");
}

void monit_runp(k5_2_octadrunp_t *runp)
// run parameter monitor
{
       printf(" ---------- RUN CONDITION ---------\n");
       printf("   k5name  : %s\n",runp->k5name);
       printf("   octadname_given : %s\n",runp->octadname_given);
       printf("   octadname       : %s\n",runp->octadname);
       printf("   outdir    : %s\n",runp->outdir);
       printf(" -------------------------------------\n");
    printf("\n");
}

int main(int argc, char *argv[])
{
       int i;
       double telapsed;           // to measure processing time
       char fname[256];
       //char fpath[256];
       char program_version[60];
	   time_t tstart, tfinish;    
       //mk5tok5_info_t mk5info;
       //gr_info_t gr; 
       k5_2_octadrunp_t runp;

       if(argc == 1 || (argc == 2 && strncmp(*(argv+1),"--help",6)==0 )){ // 2008-02-19
          print_usage();
          exit(0);
       }
       if(argc == 2 && strncmp(*(argv+1),"--ver",5)==0 ){ // 2008-02-19
          printf("\n");
          printf("%s\n",PROGRAM_VERSION);
          printf("\n");
          exit(0);
       } 
       if (argc>=2){
          if(strncmp(*(argv+1),"env",3)==0){
          printf("   Enviroment variables\n");
          printf("        %s ---   default directory for Mark5 data\n",M5DIR);
          printf("                      ( %s )\n",getenv(M5DIR));
          printf("        %s ---   default directory for Mark5 vex file\n",M5VEX);
          printf("                      ( %s )\n",getenv(M5VEX));
          printf("\n");
          exit(0);
          }
       }

       strcpy(program_version,PROGRAM_VERSION);

       /* run parameters get */
       param_get_runp(argc, argv, &runp); 

       /* out directory create if not set */

//       if(runp.outdir[0]=='\0'){
//          parse_dir_and_name(runp.octadname, runp.outdir, fname); // debug
          //printf("A %s\n",runp.mk5name);
          //printf("A %s\n",runp.outdir);
          //printf("A %s\n",fname);
//       }

	   // generate OCTAD file name
	   if(runp.octadname_given[0]=='\0'){
		   strcpy(runp.octadname,runp.k5name);
		   strcat(runp.octadname,".raw");
	   } else {
		   strcpy(runp.octadname,runp.octadname_given);
	   }


       /* Program Name Display */
       printf("************************************************************\n");
       printf("*    K5/VSSP to OCTAD Data Format Converter                *\n");
       printf("*      %s  by  T.KONDO/NICT    *\n",program_version);
       printf("************************************************************\n");
       printf("\n");

       monit_runp(&runp);

       tstart= time(NULL);


	   /* Mark5 data structure check if selected so*/
//       if(runp.ptype==1){  // mode 1 run parameters
//          k=0;
//          for(n=0; n<mk5info.ntrack; n++) k+=mk5info.bit_valid[n] ;
//          if(k == 0) {  // novalid track
//              printf(" No valid track was found in the information file.\n");
//              goto exitend;   
//          } 
//       }


//	   if(runp.k5numch==1){
//	       runp.group=0;
//		   runp.m5ch=runp.chtable[0];
//		   runp.channel=runp.m5ch;
//	   } else {
//	       runp.group=1;
//	   }


   /********************************************/
   /* loop entry of Frequency group processing */ 
   /********************************************/

//loop_g:

       i=convK5_2_OCTAD(&runp);

       if(i < 0) goto exitend;

       printf("  OCTAD file (%s) has been created.\n\n",runp.octadname);

//       if(runp.ptype==1){  // in case of Mode 1 operation
//           if(runp.channel==0 && runp.group==0) goto loop_g;
//       }

//exit_loop_g:


       tfinish= time(NULL);
       telapsed = difftime(tfinish,tstart);
       printf("Time elapsed (sec) for One obs process is %f\n",telapsed);

exitend: 
       return 0;
}

void k5_header_check(unsigned int *head,  int *k5_flag, int64_t *bytes_sec, int *seconds)
{
/*
     K5 header check

	 [in]
	    head[8]  ---  header data 

	 [return]
	    k5_flag  ---  0 :  not K5 header
		              1 :  VSSP header
					  2 :  VSSP32 header
	    bytes_sec --  # of bytes in a second (without header)
		seconds ----  time in second
	    
*/

	int flag,secs,sfreq,numch,adbit,fid,aid;
	char cw,cbuf[9];
    int sdat[16]={1,2,4,8,0,0,0,0,0,0,0,0,0,0,0,0};
    int fdat[16]={40,100,200,500,1000,2000,4000,8000,16000, 
                  32000,64000,128000,256000,512000,1024000,2048000}; 

	flag=0;
	secs=0;
	numch=0;
	adbit=0;
	sfreq=0;

    /* check 1st sync pattern */
	if(head[0]!=0xFFFFFFFF) goto exitreturn;  // not a K5 header

    /* check 2nd sync pattern */

	if(((head[1] >> 24) & 0xFF) == 0x8B) flag = 1;  // VSSP header
	if(((head[1] >> 24) & 0xFF) == 0x8C) flag = 2;  // VSSP32 header
	if(flag==0) goto exitreturn;

	/* check number of channels */
	if((head[1] & 0x00020000) == 0) numch=1;
	if((head[1] & 0x00020000) == 0x00020000) numch=4;

	/* check sampling frequency (kHz) */
	fid=(head[1] >> 18) & 0xF ;
	sfreq=fdat[fid];

	/* check adbit */
	aid=(head[1] >> 22) & 0x3;
	adbit=sdat[aid];

	/* pick up seconds */
	secs=head[1] & 0x1FFFF;

exitreturn:
	/* set return parameters */
	*bytes_sec=(int64_t)sfreq*1000*numch*adbit/8;
	*seconds=secs;
	*k5_flag=flag;
}


int convK5_2_OCTAD(k5_2_octadrunp_t *runp)
/*
    K5 to OCTAD conversion  2014-06-01 
	     (support only 1 ch mode)
*/
{
	unsigned int head[8];   // for VSSP32 header (2 X 32 bits)
	unsigned int obuf[20000];
	int nbuf=20000;
	double fsample;  // sampling frequency

    int i,n,k;
	int imax;
    int iwrite;
	int iread;
	int numch;
	int adbit;       // AD bits in a Mark5B (1 or 2)
	int k5numch;     // # of ch in a K5 format data (1 or 4)
    int ierr;
	int frame_c;   // 
	int headlen;
	int seconds;
	int k5_flag;
	int64_t bytes_sec;
    FILE *fo,*fi;

	printf("  convK5_2_OCTAD : \n");
    ierr=0;
	frame_c=0; 


	ierr=0;

	// file open
	if( (fi  = fopen( runp->k5name, "rb" )) == NULL ){
         printf("File (%s) Open Error! \n",runp->k5name);
		 ierr=-1;
         goto exit2return;
    }

	headlen=8;
    iread = fread( head, sizeof( unsigned int ), headlen, fi );


    k5_header_check(head, &k5_flag, &bytes_sec, &seconds);

	if(k5_flag==0) goto exit2return;
	if(k5_flag==1) headlen=2;

	if(((bytes_sec/4) % nbuf) !=0){
		printf(" Buffer size (%d) not match with #bytes/sec (%lld)!!\n",nbuf,(long long)bytes_sec);
		i=fclose(fi);
		ierr=-1;
        goto exit2return;
	}

	imax=(bytes_sec/4)/nbuf;

	// open output file

    if( (fo  = fopen( runp->octadname, "wb" )) == NULL ){
       printf("  File  %s  Open Error! \n",runp->octadname);
       ierr= -2;
	   i=fclose(fi);
	   goto exit2return;
    }  

    fseek(fi,0,SEEK_SET); // set reading pointer to the first

	while(1){
		// header read
        iread = fread( head, sizeof( unsigned int ), headlen, fi );
		if(iread==0) {
			goto exithere;
		}
        k5_header_check(head, &k5_flag, &bytes_sec, &seconds);
		printf(" %d\n",seconds);

		// data copy
		for(n=0; n<imax; n++){
           iread = fread( obuf, sizeof( unsigned int ), nbuf, fi );
		   if(iread==0) {
			  goto exithere;
		   }
           iwrite=fwrite(obuf,sizeof(unsigned int),iread,fo);
		}
	}
exithere:
    i=fclose( fi) ;
    // rest of data write
    i=fclose( fo ) ;
    //printf("  ****************************************************************\n");
	printf("\n");

exit2return:
       //printf("  K5 file (%s) has been created.\n\n",runp->k5name);

    return ierr;
}
