/********************************************************************/
/*                                                                  */
/*   Program Name :  speana_vdif                                    */
/*                                                                  */
/*      Version:  1.30  2018-03-08                                  */
/*                                                                  */
/*      Copyright (c) 2014-2018 T.Kondo All Right Reserved          */
/*                                                                  */
/********************************************************************/
/*

  Spectrum analysis of VDIF data
        (off-line processing)

          T.Kondo 2002-05-20
               Ver 0.00 2014-04-16 start development based on speana 
               Ver 1.00 2014-06-13 
               Ver 1.10 2016-08-03 2bit decode mode default is changed to offset binary 
               Ver 1.11 2017-02-22 update to use FFT results as defined way (time->frequency)
               Ver 1.20 2018-02-03 support # of channels exceeding 16 (MAX is 64)
               Ver 1.21 2018-02-07 add "-fs" option to set sampling frequency for VDIF not including sampling frequency
               Ver 1.30 2018-02-08 
               Ver 1.40 2018-04-10 add options '-noinfo', '-velocity baserfmhz refmhz', '-nohead', '-nohead1', '-nohead2'
                                   add members " flag_no_info", "flag_velocity", "rfmhz", "refmhz" 
                                   " flag_no_title", "mode_title" to speana_vdif_runp_t variable
                                   add option '-ps psname' for set postscript file other than pgplot.ps
               Ver 1.41 2018-04-26 add option '-lsb' to set LSB mode in case of I/Q data 


is\b17341_is1_1823_2017342182311.vdif
yj\b17341_yj_342-1822.m5a

  How to run:

    speana_vdif vdif_file [options]  
    
       where  vdif_file -- vdif file name
                           if 0 is given, internal dflt name will be used

           options 
               -m[ode] YX 
			         axis scale mode
                       X: frequency axis mode
                          =0: log scale  =1: linear scale (default)
                       Y: amplitude axis mode
                          =0: log scale (default)  
                          =1: linear scale (common for 4 channels)
                          =2: linear scale (each max and min)
			   -a[uto_corr] [width]
			         set auto correlation mode.
					 width : set display width in time axis  
					         (deflt is full span)
               -i[nteg] integration_time
                     integration period (sec). Default is 1.0
               -p[lot_device] pmode
			         set plot device
                       0: XWINDOW and PostScript file out (default)
                       1: PostScript out only 
                       2: XWINDOW only 
               -c[omment] comment
			         comment (displayed at upper side of graph)
                     If omitted, you will be asked later.
               -s[offset] soffset
			         +VE : start offset (sec) (default is 0.0)
					 -VE : absolute value means skip over bytes 
               -k[ashima] [bytesframe] 
                     set header format to Kashima compulsory 
                     bytesframe: +VE bytes/frame set compulsory
			   -2bit mode     -- set 2bit AD decode table
			                         1: offset binary   0,1,2,3 (no conversion) (default)
									 2: Mark V          0,2,1,3
									 3: signed integer  1,0,2,3
               -thread n   -- set pick up thread number (default is 1st thread number)(2018-02-02)
               -fs fsMHz   -- set sampling frequency(MHz) for VDIF not including sampling fequency

               -ps psname  --- set postscript file name other than pgplot.ps
               -noinfo  -- suppress processing date and file information on a graph
               -velocity basemhz refmhz  
                        -- set velocity axis instead of frequency axis
                           basemhz : baseband RF frequency (MHz) -VE means LSB
                           refmhz : reference frequency (MHz) where velocity=0
                                          velocity is computed as follows
                                          V=C*df/f0
                                             where C: light velocity
                                                   f0: reference frequency
                                                   df: frequency from f0 
               -nohead|-notitle  -- suppress title in a graph
               -nohead1|-notitle1 -- suppress time info in title
               -nohead2|-notitle2 -- suppress sampling and time information 

               -lsb   -- set LSB mode in case of I/Q data (default is USB mode)


*/

#define PROGRAM_VERSION "Ver. 2018-04-26"

//#define METHOD_A   // for method1    I Q doesn't consider complex     (just checking purpose)

#define METHOD_B   // for method2    I Q merely complex  (may be this is the best)  <== for normal operation

//#define METHOD_B1   // same as B but double the sampling frequency artificially   (just checking purpose)

//#define METHOD_C   // for method3    I Q combines to real data multiplying 0.5*sampling freq  (just checking purpose)

//#define METHOD_C1   // for method3    I Q combines to real data using Hilbert filtering  ( 2nd best) <== for I/Q to Real

//#define METHOD_D   // for method4    I Q combines to real data using IFFT  (just checking purpose)

//#define USE_GNUPLOT    // use GNUPLOT  (default is use of PGPLOT)

//#define DEBUG_DATA_OUT   // to output text data

#include "libvdif.h"

#include "g_head.h"    // header for graphics

#ifdef WIN32
  #define USE_FFTW    // use FFTW package
#endif

#ifdef USE_FFTW
 #ifdef WIN32
  #include "fftw3.h"     // use of FFTW package
 #else
  #include <fftw3.h>     // use of FFTW package
 #endif
#endif

//#include "libk5.h"      // common header for IP-VLBI utilities

#include "libfft.h"

#ifdef WIN32
#define USE_ASM    // use inline assemble function
#else
#define USE_ASM    // use inline assemble function
#endif
  
//#include "libvssp32.h"

/*    Offset from dB to dBm conversion from A/D multibit sampling data */ 
#define OFFSET1  16.0     //1 bit sampling full power convert to 0 dB (theoretical)
      /* following constants are provisional  (measured on 2002/5/18) */
#define OFFSET2   9.0     //2 bit sampling data offset measured at 200kHz for PC(vlbi11)
#define OFFSET4  -1.0     //4 bit sampling data offset measured at 200kHz for PC(vlbi11)
#define OFFSET8 -26.0     //8 bit sampling data offset measured at 200kHz for PC(vlbi11)

#define BUFMAX 25000   // maximum read buffer size
//#define BUFMAX 2500   // maximum read buffer size

#ifndef PI
#define PI 3.141592653
#endif

#define DFLT_DATA_FILE "vdif.dat"

#ifdef WIN32
   int tempcounter;  // counter for tempfile manage
#endif

#include "libvdif.h"
#include "libspeplot_v.h"


#define DEFLT_VDIF_FILE "vdif.dat"   // default vdif data file name

typedef struct speana_vdif_runp     // run parameters for vdifcheck
{
    char vdifname[256]; 
	char comment[60];
    char psname[128]; // PS file name (default pgplot.ps   2018-04-10
    int mode;   //  YX: axis scale mode
    double tinteg;
	double soffset;
    double f1khz;     // search frequency range (lower boundary) (kHz)
    double f2khz;     // search frequency range (upper boundary) (kHz)
	int64_t skip_bytes;   // skip bytes (1st skip over # of bytes)
    int pmode;
	int auto_flag;
	int smode;
	int flag_compulsory;
    int e2bit;
    int thread_num;   // 2018-02-03
    double fsMHz;   // sampling frequency (MHz) 2018-02-07
    int flag_no_info;        // 2018-04-09  suppres display of file name anfnd processing date
    int flag_velocity;       // 2018-04-09  display velocity instead of frequency
    int flag_no_title;       // 2018-04-10  title display flag 1:no display 0: display title
    int mode_title;       // 2018-04-10 title display mode 0: full 1: omit time info: 2: omit both sampling and time
    double rfmhz;  // 2018-04-09 base band RF frequency (MHz) (used in case of flag_velocity=1) -VE means LSB
    double refmhz; // 2018-04-09 RF frequency where set velocity = 0  (used in case of flag_velocity=1)
    int flag_lsb_IQ;    // 2018-04-26 set LSB mode in case of I/Q data
} speana_vdif_runp_t;

void calc_spectrum_from_IQ(int flag_lsb, int *idata, int *qdata, int ndata, int nfft, double *f_hanning, double *cspe, 
#ifdef USE_FFTW
                   fftw_complex *cc, fftw_plan pfcc, fftw_plan pbcc);
#else
                   double *cdata);
#endif
void decode_IQ(int flag_lsb, int *idata, int *qdata, int ndata, int *sphase, int *rdata);

void calc_normal_spectrum(int *idata, int ndata, int nfft, double *f_hanning, double *cspe, 
#ifdef USE_FFTW
                   fftw_complex *cc, fftw_plan pfcc, fftw_plan pbcc);
#else
                   double *cdata);
#endif

void iq_to_real(int flag_lsb, int *idata, int *qdata, int ndata, int *rdata);
void decode_IQ_HF(int flag_lsb, int *idata, int *qdata, int ndata, int *rdata);
void save_IQ_for_debug(int *idata, int *qdata, int ndata, int mode);   // 2018-04-30 for debug

void print_usage(void)
{
    printf("speana_vdif %s\n",PROGRAM_VERSION);
    printf("\n");
    printf("Usage: speana_vdif vdif_file [options]\n");
    printf("\n");
    printf("       where  vdif_file -- vdif file name\n");
    printf("                           if 0 is given, internal dflt name will be used\n");
    printf("\n");
    printf("          options \n");
    printf("               -m[ode] YX \n");
    printf("                     axis scale mode\n");
    printf("                       X: frequency axis mode\n");
    printf("                          =0: log scale  =1: linear scale (default)\n");
    printf("                       Y: amplitude axis mode\n");
    printf("                          =0: log scale (default)  \n");
    printf("                          =1: linear scale (common for 4 channels)\n");
    printf("                          =2: linear scale (each max and min)\n");
    printf("               -a[uto_corr] [width]\n");
    printf("                     set auto correlation mode.\n");
    printf("                       width : set display width in time axis \n");
    printf("                               (deflt is full span)\n");
    printf("                     at the peak is displayed (magnified)\n");
    printf("               -i[nteg] integration_time\n");
    printf("                     integration period (sec). Default is 1.0\n");
    printf("               -p[lot_device] pmode\n");
    printf("                     set plot device\n");
    printf("                       0: XWINDOW and PostScript file out (default)\n");
    printf("                       1: PostScript out only \n");
    printf("                       2: XWINDOW only \n");
    printf("               -c[omment] comment\n");
    printf("                     comment (displayed at upper side of graph)\n");
    printf("                     If omitted, you will be asked later.\n");
    printf("               -s[offset] soffset\n");
    printf("                     +VE : start offset (sec) (default is 0.0)\n");
    printf("                     -VE : absolute value means skip over bytes \n");
    printf("               -k[ashima] [bytesframe]\n");
    printf("                     set header format to Kashima compulsory \n");
	printf("                     bytesframe: +VE is bytes/frame set compulsory\n");
    printf("               -2bit mode -- set 2bit AD decode table\n");
    printf("                           1: offset binary   0,1,2,3 (no conversion)(default)\n");
    printf("                           2: Mark V standard 0,2,1,3\n");
    printf("                           3: signed integer  1,0,2,3\n");
    printf("               -thread n   -- set pick up thread number (default: 1st thread number)\n");
    printf("               -fs fsMHz   -- set sampling frequency(MHz) for VDIF not including sampling fequency\n");
    printf("               -ps psname  --- set postscript file name other than pgplot.ps\n");
    printf("               -lsb   -- set LSB mode in case of I/Q data (default is USB mode)\n");
    printf("               -noinfo  -- suppress processing date and file information on a graph\n");
    printf("               -velocity baserfmhz refmhz\n");
    printf("                        -- set velocity axis instead of frequency axis\n");
    printf("                           baserfmhz : baseband RF frequency (MHz) -VE means LSB\n");
    printf("                           refmhz : reference frequency (MHz) where velocity=0\n");
    printf("               -nohead | -notitle  -- suppress title in a graph\n");
    printf("               -nohead1 | -notitle1 -- suppress time info in title\n");
    printf("               -nohead2 | -notitle2 -- suppress sampling and time information \n");
    printf("\n");
}

void param_get_speana_vdif(int argc, char *argv[], speana_vdif_runp_t *runp)                       
{
    char cbuf[256];
    char cbuf2[50];

    // initialize parameters

    runp->mode=1;
	runp->auto_flag=0;   // auto correlation flag 
    runp->pmode=0;
    runp->vdifname[0]='\0';
	runp->comment[0]='\0';
	runp->tinteg=1.0;
	runp->soffset=0.0;
	runp->flag_compulsory=0;
	runp->skip_bytes=0;
    //runp->e2bit=2; // default
    runp->e2bit=1; // default  from 2016-08-03
    runp->thread_num=-1; // 2018-02-03
    runp->fsMHz=0.0;
    runp->f1khz=0.0;
    runp->f2khz=0.0;
    strcpy(runp->psname,PS_OUT_FILE);   // 2018-04-10

    runp->flag_no_info=0;  // 2018-04-09
    runp->flag_velocity=0; // 2018-04-09
    runp->rfmhz=0.0;       // 2018-04-09
    runp->refmhz=0.0;      // 2018-04-09
    runp->flag_no_title=0;   // 2018-04-10
    runp->mode_title=0;      // 2018-04-10
    runp->flag_lsb_IQ=0;   // 2018-04-26

    /* 1st parameter (vdif file name) get */
    if (argc>=2){
       strcpy(runp->vdifname,*(argv+1));
	   if (strlen(*(argv+1))==1 && runp->vdifname[0]=='0'){  // set default name
          strcpy(runp->vdifname,DEFLT_VDIF_FILE);
	   }
    }

    /* command parameters get */
    if(cmnd_get(argc,argv,"-m",cbuf)==0) runp->mode = atoi(cbuf);      // mode
    if(cmnd_get(argc,argv,"-i",cbuf)==0) runp->tinteg = atof(cbuf);    // integration
    if(cmnd_get(argc,argv,"-p",cbuf)==0) runp->pmode = atoi(cbuf);     // plot device
    if(cmnd_get(argc,argv,"-c",cbuf)==0) strcpy(runp->comment,cbuf);   // comment
	if(cmnd_get(argc,argv,"-s",cbuf)==0){
		if(atof(cbuf) >= 0.0){
		    runp->soffset = atof(cbuf);   // second offset
		} else {
			runp->skip_bytes=-atoi(cbuf);  // skip bytes
		}
	}
    if(cmnd_get(argc,argv,"-a",cbuf)==0){                              // auto correlation mode
		runp->auto_flag=1;
        if(cbuf[0] != '\0'){
            if(cbuf[0] != '-') runp->auto_flag=atoi(cbuf);
        }
    }
    if(cmnd_get(argc,argv,"-k",cbuf)==0 ||cmnd_get(argc,argv,"-K",cbuf)==0 ){ // set compulsory mode
		runp->flag_compulsory=1;
        if(cbuf[0] != '\0'){
            if(cbuf[0] != '-') runp->flag_compulsory=atoi(cbuf);
        }
    }
	if(cmnd_get(argc,argv,"-2bit",cbuf)==0) runp->e2bit = atoi(cbuf);   // 2bit encode mode
    //if(cmnd_get(argc,argv,"-thread",cbuf)==0) runp->thread_num = atoi(cbuf);   // thread number
    if(cmnd_get(argc,argv,"-t",cbuf)==0) runp->thread_num = atoi(cbuf);   // thread number
    if(cmnd_get(argc,argv,"-fs",cbuf)==0) runp->fsMHz = atof(cbuf);   // sampling frequency (MHz)

    if(cmnd_get(argc,argv,"-ps",cbuf)==0) {     // -ps psname    2018-02-19
        if(cbuf[0] != '\0'){
            strcpy(runp->psname,cbuf); 
        }
    }
    if(cmnd_get(argc,argv,"-noin",cbuf)==0){                     // 2018-04-09
        runp->flag_no_info=1;
    }
    if(cmnd_get_n2para(argc, argv, "-vel", cbuf, cbuf2)==0){      // 2018-04-09
        runp->flag_velocity=1;
        runp->rfmhz=atof(cbuf);
        runp->refmhz=atof(cbuf2);
    }
    if(cmnd_get_np(argc,argv,"-nohead",cbuf)==0){                     // 2018-04-10
        runp->flag_no_title=1;
    }
    if(cmnd_get_np(argc,argv,"-nohead1",cbuf)==0){                     // 2018-04-10
        runp->mode_title=1;
    }
    if(cmnd_get_np(argc,argv,"-nohead2",cbuf)==0){                     // 2018-04-10
        runp->mode_title=2;
    }
    if(cmnd_get_np(argc,argv,"-notitle",cbuf)==0){                     // 2018-04-10
        runp->flag_no_title=1;
    }
    if(cmnd_get_np(argc,argv,"-notitle1",cbuf)==0){                     // 2018-04-10
        runp->mode_title=1;
    }
    if(cmnd_get_np(argc,argv,"-notitle2",cbuf)==0){                     // 2018-04-10
        runp->mode_title=2;
    }
    if(cmnd_get(argc,argv,"-lsb",cbuf)==0) runp->flag_lsb_IQ=1;   // 2018-04-26 LSB mode in case of I/Q data


}



int main(int argc, char *argv[])
{
    int nloop;
    int i,n,m, k, kaisu;
    int k_at_max;  // 2018-03-05
    double freq_at_max; // 2018-03-05
    int w;
    int nfft;
    int ymode; // amplitude scale mode   0: log   1: linear
    //int scount[16];  // allow max 16 channel
    int scount[64];  // allow max 16 channel 2018-02-03
	//int upos,ss; // for debug
	int r_thread_id;
    double fact;
    double *cdata;
    double *f_hanning;
    double theta;
    double sum,dc;
    float pows;
    float pmax,pmin,pmaxdbm,pmindbm,cofst;
    float ymaxdb,ymindb;
#ifdef USE_FFTW
    fftw_complex *cc;
    fftw_plan pfcc, pbcc;
#endif
    double nfact;
    double tinteg;
    double telapsed;           // to measure processing time
    int flag_complex;  // 2018-02-09
    int flag_lsb;  // 2018-04-26
    int *idata, *qdata;  // 2018-02-09
    double *cspe;        // 2018-02-09
    int ndata;           // 2018-02-09
    int ndata2;          // 2018-03-01
    //int sphase;          // 2018-02-11
    int sphase[64];          // 2018-03-02
    int smphase;          // 2018-02-11
    int *rdata;           // 2018-02-11
    clock_t cstart1sec, cfinish1sec;  // to calculate processing time
    char ltime1st[60];
    char pgdevice_dflt[20];
    char* dflt_fname=DFLT_DATA_FILE;
    speana_vdif_runp_t runp;
	speanav_param_t st;
    vdif_file_t vd;
    timeval_vdif_t time1st;
    int *re_conv2bit;
    int re_conv2bit_1[]={0,1,2,3};   // offset binary (no conversion 2 bit)(default)
    int re_conv2bit_2[]={0,2,1,3};   //  conversion 2 bit
    int re_conv2bit_3[]={1,0,2,3};   //  conversion 2 bit singed integer



    if(argc == 1 || (argc == 2 && strncmp(*(argv+1),"--help",6)==0 )){ // 2008-02-19
          printf("speana_vdif %s",PROGRAM_VERSION);
#ifdef USE_FFTW
          printf(" compliled with FFTW3.0 \n");
#endif
#ifdef USE_GNUPLOT   /* using gnuplot */
          printf(" compliled for GNUPLOT \n");
#else
          printf(" compliled for PGPLOT \n");
#endif
          print_usage();

#ifndef USE_GNUPLOT   // using gnuplot
          printf("\n");
          printf("  Note: %s ---  environment variable for PGPLOT display device\n",PGDISP);
          printf("                   = %s",getenv(PGDISP));
          if(getenv(PGDISP)==NULL) {
              printf(",  program deflt is ( %s )\n",PGDISP_DFLT);
          } else {
              printf("\n");
          }
#endif
          printf("\n");
          exit(0);
    } 
    if(argc == 2 && strncmp(*(argv+1),"--ver",5)==0 ){
          printf("speana_vdif %s",PROGRAM_VERSION);
#ifdef USE_FFTW
          printf(" compliled with FFTW3.0 \n");
#endif
#ifdef USE_GNUPLOT   /* using gnuplot */
          printf(" compliled for GNUPLOT \n");
#else
          printf(" compliled for PGPLOT \n");
#endif
          printf("\n");
          exit(0);
    }

    param_get_speana_vdif(argc, argv, &runp);

    /* set graphic device */
    strcpy(st.ps_fname,runp.psname);   // 2018-04-10
#ifdef USE_GNUPLOT   // using gnuplot
    strcpy(st.g_display,G_DISPLAY);
    strcpy(pgdevice_dflt,G_DISPLAY);
    strcpy(st.g_ps_device,G_PS_DEVICE);
#else
    strcpy(st.g_display,select_env_dflt(PGDISP, PGDISP_DFLT));
    strcpy(pgdevice_dflt,st.g_display);
    sprintf(st.g_ps_device,"%s%s",st.ps_fname,G_PS_DEVICE);
#endif

    if(runp.e2bit==1){
       re_conv2bit=re_conv2bit_1;
    } else if(runp.e2bit==3){
       re_conv2bit=re_conv2bit_3;
    } else {
       re_conv2bit=re_conv2bit_2;
    }

	strcpy(st.fname,runp.vdifname);
	strcpy(vd.fname,runp.vdifname);
	strcpy(st.comment,runp.comment);
	st.soffset=runp.soffset;


    // plot mode set
	st.mode2=runp.mode;
	st.mode=1;  // default frequency axis mode (1= linear scale)
	ymode=0;    // dfault amplitude axis mode (= log scale)
	st.amode=runp.auto_flag;  // auto correlation mode
//	if(st.mode2 <0 ){
//        st.amode=-st.mode2;
//        st.mode=0;
//    } else {
//        st.amode=0;
//        ymode=st.mode2/10;
//        st.mode=st.mode2 % 10;
//	}

	st.sekibun = runp.tinteg;
	st.pmode = runp.pmode;
    st.f1khz=runp.f1khz;
    st.f2khz=runp.f2khz;

    st.flag_no_info=runp.flag_no_info;   // 2018-04-09
    st.flag_velocity=runp.flag_velocity; // 2018-04-09
    st.rfmhz=runp.rfmhz;                 // 2018-04-09
    st.refmhz=runp.refmhz;               // 2018-04-09
    st.flag_no_title=runp.flag_no_title;  // 2018-04-10
    st.mode_title=runp.mode_title;        // 2018-04-10

    st.flag_lsb=runp.flag_lsb_IQ;   // 2018-04-26
    flag_lsb=runp.flag_lsb_IQ;      // 2018-04-26


    printf("speana_vdif %s",PROGRAM_VERSION);
#ifdef USE_FFTW
    printf(" compliled with FFTW3.0 \n");
#else
    printf("\n");
#endif

    cstart1sec = clock();

    /* get comment */

//	if (st.comment[0]=='\0'){
//       /* enter any comment */
//       printf("Enter any comment -> ");
//       read1line(st.comment);      // 1 line read from stdin
//    }

	vd.skip_bytes=runp.skip_bytes;
	vd.flag_compulsory=runp.flag_compulsory;

    //if(vdif_t_ini(&vd,0)<0) goto exitend2;   // VDIF data check with auto mode
    if(vdif_t_ini(&vd,0)<0) exit(0);   // VDIF data check with auto mode
    //if(vdif_t_ini(&vd,1)<0) goto exitend2;  // VDIF data check with 1 sec reading compulsory
    if(vd.vhead.srate == 0){
        if(runp.fsMHz <= 0.0){
            printf(" This VDIF file does not include sampling frequency information!\n");
            printf(" Please set sampling frequency using \"-fsampl fmhz\" option.\n");
            exit(0);
        }
    }

	time1st=vd.vhead.t;  // save 1st time 

    if(runp.fsMHz != 0.0) vd.fsample=1.0e6*runp.fsMHz;  // 2018-02-07
    if(vd.fsample==0.0){
        if(runp.fsMHz==0.0){
            printf("Warning!  sampling frequency is not included in VDIF headder!!\n");
            printf("Please set sampling frequency using \"-fsampl fMHz\" option!!\n");
            exit(0);
        } else {
            vd.fsample=1.0e6*runp.fsMHz;
        }
    } else {
        if(runp.fsMHz != 0.0){  // compulsory set sampling frequency  and ndramesec
            vd.fsample=1.0e6*runp.fsMHz;
            vd.nframsec=(int)(1.0/((double)vd.frame_bytes*8.0/((double)(vd.adbit*vd.numch)*vd.fsample))+0.5); 
        }
    }

    /* parameter set */
	st.frames_per_sec=vd.nframsec;
	st.adbit=vd.adbit;   // AD bits
	st.sfreq=vd.fsample; 
	st.numch=vd.numch;
	
	/* 
	  set reading buffer size to satisfy following condition
	     1. (frame size)/(buffer size) should be an integer number
	*/
	vd.nword=vd.frame_bytes/4;     
	while (vd.nword > BUFMAX) vd.nword/=2;

	if(((vd.frame_bytes/4) % vd.nword) != 0){
		printf(" speana_vdif: buffer size does\'nt satisfy the condition! %d %d\n",vd.nword, vd.frame_bytes*4);
	}

	vd.imax=(vd.frame_bytes/4)/vd.nword;   // # ofloops to read entire one frame data
	vd.nsample=vd.nword*32/vd.adbit;  // # of samples in buffer (don't care about # of Chs)

	vd.imax_c=0;

	runp.smode=1;   // mode to choose unit processing sample numbers per channel
	vd.smode=runp.smode;
	/* set processing unit number of samples per channel */
    if(vd.smode ==0){ vd.usampl=200000; }
    else if (vd.smode ==1) {
       vd.usampl=10000;
    } else {
       vd.usampl=1000;      // for smode=2
    }
    //    check usampl condition 
    //    because usampl*adbit*numch/32 must be an integer (<= this comes from K5 fx_cor)
	//    if data is not used for K5 fx_cor, this condition can be ignored.
    if((vd.usampl*vd.adbit*vd.numch)%32 != 0) {
        while ((vd.usampl*vd.adbit*vd.numch)%32 != 0) {
             vd.usampl*=2;
        }
		printf(" new usampl %d\n",vd.usampl);
    }

	//if(vd.usampl*vd.numch > vd.nsample) vd.usampl=vd.nsample/vd.numch;

	vd.unitp=(double)vd.usampl/vd.fsample;   // period of usampl


	//printf("imax nword32  %d %d\n",vd.imax,vd.nword);    //debug

    nfft=1024;
    if(vd.usampl < 524288) nfft= 524288;
    if(vd.usampl < 262144) nfft= 262144;
    if(vd.usampl < 131072) nfft= 131072;
    if(vd.usampl < 65536)  nfft= 65536;
    if(vd.usampl < 32768)  nfft= 32768;
    if(vd.usampl < 16384)  nfft= 16384;

    st.nfft=nfft;

    //nfact=2.0/(double)nfft;
    nfact=1.0/(double)nfft;  // 2017-02-22

    vd.ubuf  = (u_int32_t *)malloc(vd.nword*sizeof(u_int32_t)); //  data read buffer
	//vd.idata  = (int *)malloc((vd.usampl+32)*vd.numch*sizeof(int)); //  data pick up buffer (32 is overflow space)
    vd.ch   = (ch_data_t *)malloc(vd.numch*sizeof(ch_data_t));
    for(n=0; n<vd.numch; n++)
    {
        vd.ch[n].idata=(int *)malloc((vd.usampl+32)*vd.numch*sizeof(int)); //  data pick up buffer (32 is overflow space)
    }

    idata=(int *)malloc((vd.usampl+32)*sizeof(int));  // 2018-02-09
    qdata=(int *)malloc((vd.usampl+32)*sizeof(int));  // 2018-02-09

    rdata=(int *)malloc((vd.usampl+32)*sizeof(int));  // 2018-02-11

#ifdef USE_FFTW
    cc=fftw_malloc(sizeof(fftw_complex)*nfft);

    pfcc=fftw_plan_dft_1d(nfft, cc, cc, FFTW_FORWARD, FFTW_ESTIMATE);
    pbcc=fftw_plan_dft_1d(nfft, cc, cc, FFTW_BACKWARD, FFTW_ESTIMATE);
#endif

    cdata = (double *)malloc(2*nfft*sizeof(double));
    cspe = (double *)malloc(2*nfft*sizeof(double));   // 2018-02-09
    //st.spows = (float *)malloc(4*nfft*sizeof(float));
    st.spows = (float *)malloc(vd.numch*nfft*sizeof(float));
    f_hanning = (double *)malloc(nfft*sizeof(double));

    //Hanning filter coefficients prepare
    for(k=0; k<nfft; k++){
        theta=360.0*(double)k/(double)(nfft-1) - 180.0;
        theta=fmod(theta,360.0) ;
        f_hanning[k]=0.5*(1.0+cos(theta*PIRA));
    }

    //fact=((float)vd.usampl)*((float)vd.usampl);
    fact=(double)nfft/(double)vd.usampl;  // 2017-02-22
    fact*=fact;                           // 2017-02-22


    for(k=0; k<vd.numch*nfft; k++) st.spows[k]=0.0;
    for(k=0; k<vd.numch; k++) scount[k]=0;
    for(k=0; k<vd.numch; k++) sphase[k]=0;  // 2018-03-02

	// file open
	if( (vd.fin  = fopen( vd.fname, "rb" )) == NULL ){
         printf("File (%s) Open Error! \n",vd.fname);
         goto exitend2;
    }
	vd.flag_open=1;

	if(vd.flag_vdif_anal==1){ // read pointer move to 1st valid header
		fseek(vd.fin,vd.pos_1st_header,SEEK_SET); // move pointer by K5->bytesec
	}

    /* foward startoffset seconds */
	if(st.soffset>0){
		printf("  %d second forwading.....",(int)st.soffset);
	    if(fwdNsecVDIF((int)st.soffset, &vd) <0){
           printf("Foward error %d\n",(int)st.soffset);
           i=fclose( vd.fin ) ;
           goto exitend;
        }
	    printf("\n");
	    /* get headder info again */
        if(vdif_header_read_mode1(&vd)<0) goto exitend;
	    time1st=vd.vhead.t;
	}

	sprintf(ltime1st,"Time(UT) %04d/%02d/%02d %02d:%02d:%02d",
		time1st.year,time1st.month,time1st.mday,time1st.hh,time1st.mm,time1st.ss);
    //sprintf(ltime1st,"Debugd");  // debug
    strcpy(st.lab2,ltime1st);
    strcat(st.lab2,"  ");
    strcat(st.lab2,st.comment);

	// vdif read 

	//r_thread_id=vd.vhead.thread_id;  // set reading thread ID to 1st one
    vd.pup_thread=vd.vhead.thread_id;  // set reading thread ID to 1st one  2018-02-09

    //if(runp.thread_num >=0) r_thread_id=runp.thread_num;  // 2018-02-03
    if(runp.thread_num >=0) vd.pup_thread=runp.thread_num;  // 2018-02-09

    printf(" Pick Up thread # ==> %d\n",vd.pup_thread); // 2018-02-03

    //r_thread_id=2;
        
    tinteg=0.0;
	vd.pos=0;

	//st.numch=1; // assuming 1 ch data

    nloop=0;

	vd.tot_frame=0; // total frame counter
	vd.upos=0;
	vd.usampl_c=0;
	vd.n_usampl=0;
	vd.flag_read_header=1;
	vd.flag_read_data=1;
	vd.flag_idata_over=0;

    printf("  Reading thread ID is  %d\n",vd.pup_thread);
    printf("  Data reading.....\n");


    //sphase=0;

loop1:

	    //upos=vd.upos;    // debug
		//ss=vd.vhead.t.ss; // debug

        i=vdif_read_unit(&vd, vd.pup_thread);

        st.flag_complex=vd.vhead.d_type;    //   1: I/Q data 0: real data  2018-04-26
        flag_complex=vd.vhead.d_type;

        //flag_complex=0;
        ndata=vd.npos;
        if(flag_complex==1){ // then complex data  2018-02-09
            //flag_complex=1;
            if((vd.npos % 2) != 0) printf("Warning!!  # of data does not match with complex type data!!\n");
            ndata=vd.npos/2;
#ifdef METHOD_B1
            ndata=vd.npos;
#endif
        }
		//printf("%02d:%02d:%02d %lld %f\n",vd.vhead.t.hh,vd.vhead.t.mm,vd.vhead.t.ss,vd.tot_frame,vd.dtime); //debug

		//debug
		//if(vd.vhead.t.ss != ss){  // debug
		//  printf("%02d:%02d:%02d %lld %f\n",vd.vhead.t.hh,vd.vhead.t.mm,vd.vhead.t.ss,vd.tot_frame,vd.dtime); //debug
		//	printf("upos %d  %d  \n",upos,vd.upos);
		//}

		if(i==-99){ // EOF
			printf("\n");
			goto exitloop; 
		} else {
			if(i < 0){
			  printf("\n");
			  goto exitend; 
			}
		}

		for(m=0; m<vd.numch; m++){
		//for(m=0; 8; m++){
		    //m=0;
            kaisu=0;
            sum=0.0;

            smphase=sphase[m];  // 2018-03-02
            // initialize array
//#ifdef USE_FFTW
//            for(k=0; k<nfft; k++){ cc[k][0]=0.0; cc[k][1]=0.0;}
//#else
//            for(k=0; k<2*nfft; k++) cdata[k]=0.0;
//#endif

#ifdef METHOD_A
            for(k=0; k< vd.npos; k++) idata[k]=0;
#else
            for(k=0; k<ndata; k++){
                idata[k]=0;
                qdata[k]=0;
            }
#endif


			//for(k=0; k<vd.npos; k++){
			for(k=0; k<ndata; k++){

                /* set data */
                if(flag_complex == 1){
                   if(st.adbit==2){
               //        printf("Debug %d %d %d\n",k,vd.ch[m].idata[2*k],vd.ch[m].idata[2*k+1]);  // debug
                       //w=re_conv2bit[vd.ch[m].idata[k]];
#ifdef METHOD_B1
                       idata[k]=re_conv2bit[vd.ch[m].idata[2*(k/2)]];
                       qdata[k]=re_conv2bit[vd.ch[m].idata[2*(k/2)+1]];
#else
                       idata[k]=re_conv2bit[vd.ch[m].idata[2*k]];
                       qdata[k]=re_conv2bit[vd.ch[m].idata[2*k+1]];
#endif
                   } else {
#ifdef METHOD_B1
                       idata[k]=vd.ch[m].idata[2*(k/2)];
                       qdata[k]=vd.ch[m].idata[2*(k/2)+1];
#else
                       idata[k]=vd.ch[m].idata[2*k];
                       qdata[k]=vd.ch[m].idata[2*k+1];
#endif
                   }

                } else {
                   if(st.adbit==2){
                       idata[k]=re_conv2bit[vd.ch[m].idata[k]];

                   } else {
                       idata[k]=vd.ch[m].idata[k];
                   }
                }
            }

#ifdef DEBUG_DATA_OUT   // to output text data
            if( m==1) save_IQ_for_debug(idata, qdata, ndata, nloop);  //for debug
#endif


            if(flag_complex==1){

#ifdef METHOD_D
               iq_to_real(flag_lsb, idata, qdata, ndata, rdata);

   #ifdef USE_FFTW
               calc_normal_spectrum(rdata, 2*ndata, nfft, f_hanning, cspe, cc, pfcc, pbcc);                 
   #else
               calc_normal_spectrum(rdata, 2*ndata, nfft, f_hanning, cspe, cdata);
   #endif

#endif

#ifdef METHOD_C
               decode_IQ(flag_lsb, idata, qdata, ndata, &smphase, rdata);


   #ifdef USE_FFTW
               calc_normal_spectrum(rdata, 2*ndata, nfft, f_hanning, cspe, cc, pfcc, pbcc);                 
   #else
               calc_normal_spectrum(rdata, 2*ndata, nfft, f_hanning, cspe, cdata);
   #endif

#endif


#ifdef METHOD_C1
               decode_IQ_HF(flag_lsb, idata, qdata, ndata, rdata);


   #ifdef USE_FFTW
               calc_normal_spectrum(rdata, 2*ndata, nfft, f_hanning, cspe, cc, pfcc, pbcc);                 
   #else
               calc_normal_spectrum(rdata, 2*ndata, nfft, f_hanning, cspe, cdata);
   #endif

#endif

#ifdef METHOD_A
			for(k=0; k<vd.npos; k++){
                if(st.adbit==2){
                    idata[k]=re_conv2bit[vd.ch[m].idata[k]];
                } else {
                    idata[k]=vd.ch[m].idata[k];
                }
            }


   #ifdef USE_FFTW
               calc_normal_spectrum(idata, ndata, nfft, f_hanning, cspe, cc, pfcc, pbcc);                 
   #else
               calc_normal_spectrum(idata, ndata, nfft, f_hanning, cspe, cdata);
   #endif

#endif

#ifdef METHOD_B


   #ifdef USE_FFTW
               calc_spectrum_from_IQ(flag_lsb, idata, qdata, ndata, nfft, f_hanning, cspe, cc, pfcc, pbcc);                 
   #else
               calc_spectrum_from_IQ(flag_lsb, idata, qdata, ndata, nfft, f_hanning, cspe, cdata);
   #endif

#endif

#ifdef METHOD_B1


   #ifdef USE_FFTW
               calc_spectrum_from_IQ(flag_lsb, idata, qdata, ndata, nfft, f_hanning, cspe, cc, pfcc, pbcc);                 
   #else
               calc_spectrum_from_IQ(flag_lsb, idata, qdata, ndata, nfft, f_hanning, cspe, cdata);
   #endif

#endif


            } else {
#ifdef USE_FFTW
               calc_normal_spectrum(idata, ndata, nfft, f_hanning, cspe, cc, pfcc, pbcc);                 
#else
               calc_normal_spectrum(idata, ndata, nfft, f_hanning, cspe, cdata);
#endif
            }


            for(k=0; k<nfft; k++){
               st.spows[k*vd.numch+m]+=(cspe[2*k]*cspe[2*k]+cspe[2*k+1]*cspe[2*k+1]);
            }
            scount[m]++;
            //if(m==0) sphase=smphase;
            sphase[m]=smphase;  // 2018-03-02
        }
		tinteg+=vd.unitp;
        if(tinteg >= st.sekibun) goto exitloop;
        nloop++;

        goto loop1;

exitloop:
    if(vd.flag_open==1){
		i=fclose( vd.fin ) ;
		vd.flag_open=0;
	}

	printf("\n");

    if(st.amode > 0) {   // auto correlation function calc
       for(m=0; m<vd.numch; m++){
          for(k=0; k<nfft; k++){
             st.spows[k*vd.numch+m]/=scount[m]; // nomalization
#ifdef USE_FFTW
             cc[k][0]=st.spows[k*vd.numch+m]; cc[k][1]=0.0;
#else
             cdata[2*k]=st.spows[k*vd.numch+m];
             cdata[2*k+1]=0.0;
#endif
          }
          if(m==1 || m==3){
#ifdef USE_FFTW
              cc[0][0]=1.0;
#else
              cdata[0]=1.0;  // to avoid math error
#endif
          }
#ifdef USE_FFTW
          fftw_execute(pbcc);
#else
          cdft(2*nfft,1,cdata);
#endif
          for(k=0; k<nfft; k++){
             // if(k<10) printf("%d %e\n",k,cdata[2*k]);  // debug
             // if(k>nfft-10) printf("%d %e\n",k,cdata[2*k]);  // debug
#ifdef USE_FFTW
             if(k==0) pmax=(float)cc[0][0];
             if(cc[k][0] > pmax) pmax=(float)cc[k][0];
             st.spows[k*vd.numch+m]=(float)cc[k][0];
#else
             if(k==0) pmax=(float)cdata[0];
             if(cdata[2*k] > pmax) pmax=(float)cdata[2*k];
             st.spows[k*vd.numch+m]=(float)cdata[2*k];
#endif
          }
          for(k=0; k<nfft; k++){  // normalize
             st.spows[k*vd.numch+m]/=pmax;
             // if(k<10) printf("   %d %e\n",k,spows[k*4+m]);  // debug
             // if(k>nfft-10) printf("   %d %e\n",k,spows[k*4+m]);  // debug
          }
             
       }
         
    } else {

        /* convert power to dB  then dBm */       
        cofst=OFFSET1;       //1 bit sampling full power convert to 0 dB
        if( st.adbit==8) cofst=OFFSET8;     // 8bit sampling offset
        if( st.adbit==4) cofst=OFFSET4;     // 4bit sampling offset
        if( st.adbit==2) cofst=OFFSET2;     // 4bit sampling offset
        for(m=0; m<st.numch; m++){
            for(k=0; k<nfft; k++){
                st.spows[k*st.numch+m]/=scount[m]; // normalization
                st.spows[k*st.numch+m]=(float)(10.0*log10(st.spows[k*st.numch+m]) + cofst); 
                if(k==0){
                    pmaxdbm=st.spows[k*st.numch+m];
                    pmindbm=st.spows[k*st.numch+m];
                    if(m==0){
                        ymaxdb=pmaxdbm;
                        ymindb=pmindbm;
                    }
                    k_at_max=0;
                } else {
                    if(st.spows[k*st.numch+m] > pmaxdbm){
                        pmaxdbm=st.spows[k*st.numch+m]; 
                        k_at_max=k;   // 2018-03-05
                    }
                    if(st.spows[k*st.numch+m] < pmindbm) pmindbm=st.spows[k*st.numch+m]; 
                }
                // debug 
                //if(k>0){
                //    if(fabs(st.spows[k*st.numch+m]-st.spows[(k-1)*st.numch+m])>1.5){
                //        freq_at_max=0.5*st.sfreq*(double)k/(double)nfft;  // 2018-03-05
                //        printf("Debug peak freq (MHz) at %f\n",freq_at_max/1.0e6);
                //    }
                //}
            }
            pmax=pmaxdbm-cofst;
            pmin=pmindbm-cofst;
            //printf("CH# %d: Maximum data (dB, dBm) is %f %f\n", m+1,pmax,pmaxdbm);
#ifdef METHOD_B
            freq_at_max=0.5*st.sfreq*(double)k_at_max/(double)nfft;  // 2018-03-05
            printf("CH# %d: Maximum data (dB, dBm) is %f %f at %f MHz\n", m+1,pmax,pmaxdbm,freq_at_max/1.0e6); 
#else
            printf("CH# %d: Maximum data (dB, dBm) is %f %f\n", m+1,pmax,pmaxdbm);

#endif
            if(pmaxdbm > ymaxdb) ymaxdb=pmaxdbm;
            if(pmindbm < ymindb) ymindb=pmindbm;
        }
        printf("All CH: Maximum and Minimum data (dBm) are %f %f\n", ymaxdb, ymindb);      

    }
    cfinish1sec = clock();
    telapsed = (double)(cfinish1sec - cstart1sec) / CLOCKS_PER_SEC;
    printf("Time elapsed for processing is %f sec\n",telapsed);

#ifdef METHOD_B
    if(flag_complex==1){ // rearrange data // 2018-02-11
        st.nfft*=2;
        st.wpows = (float *)malloc(vd.numch*st.nfft*sizeof(float));    // 2018-02-11
        for(m=0; m<st.numch; m++){
            for(k=0; k<nfft; k++){
                st.wpows[k*st.numch+m]=st.spows[k*st.numch+m];
                //if(k > 0){
                //   st.wpows[(st.nfft-k)*st.numch+m]=st.spows[k*st.numch+m];
                //}
            //for(k=0; k<nfft/2; k++){
            //    st.wpows[k*st.numch+m]=st.spows[(k+nfft/2)*st.numch+m];
            //    //st.wpows[k*st.numch+m]=st.spows[(nfft-k-1)*st.numch+m];
            //    if(k > 0){
            //       st.wpows[(k+nfft/2)*st.numch+m]=st.spows[k*st.numch+m];
            //    }
            }
        }
        free(st.spows);
        st.spows=st.wpows;
    }
#endif

#ifdef METHOD_B1
    /*
    if(flag_complex==1){ // rearrange data // 2018-02-11
        st.nfft*=2;
        st.wpows = (float *)malloc(vd.numch*st.nfft*sizeof(float));    // 2018-02-11
        for(m=0; m<st.numch; m++){
            for(k=0; k<nfft; k++){
                st.wpows[k*st.numch+m]=st.spows[k*st.numch+m];
                if(k > 0){
                   st.wpows[(st.nfft-k)*st.numch+m]=st.spows[k*st.numch+m];
                }
            }
        }
        free(st.spows);
        st.spows=st.wpows;
    }
    */
#endif

    /* plot spectrum (or auto cor) POSTSCRIPT OUT and XTERM*/
     plotspe_v(&st,ymaxdb, ymindb);

  exitend: ;
    if(vd.flag_open==1){
		i=fclose( vd.fin ) ;
		vd.flag_open=0;
	}
      
    free(vd.ubuf);
    //free(vd.idata);
    for(n=0; n<vd.numch; n++) free(vd.ch[n].idata);
    free(vd.ch);
    free(cdata);
    free(st.spows);

    free(idata);    // 2018-02-11
    free(qdata);    // 2018-02-11
    free(cspe);    // 2018-02-11

#ifdef USE_FFTW
    fftw_destroy_plan(pfcc);
    fftw_destroy_plan(pbcc);
    fftw_free(cc);
#endif

 exitend2: ;

    return 0;
}





void calc_spectrum_from_IQ(int flag_lsb, int *idata, int *qdata, int ndata, int nfft, double *f_hanning, double *cspe,
#ifdef USE_FFTW
                   fftw_complex *cc, fftw_plan pfcc, fftw_plan pbcc)
#else
                   double *cdata)
#endif
{
 /*
    compute combined spectrum of LSB and USB from I and Q time siries data

           2018-02-09
 */



    int k;
    int kaisu;
    double sumi,sumq,dci,dcq;
    double nfact,fact,pows;
    double sign;

    nfact=1.0/(double)nfft;
    fact=(double)nfft/(double)ndata;   // check!!  use of ndata instead of usampl is OK?
    //fact*=fact;

    sign=1.0;
    if(flag_lsb == 1) sign=-1.0;    // LSB


    kaisu=0;
    sumi=0.0;
    sumq=0.0;

    // debug
    //printf("I-DATA");
    //for(k=0; k<100; k++) printf(" %d",idata[k]);
    //printf("\n");
    //printf("Q-DATA");
    //for(k=0; k<100; k++) printf(" %d",qdata[k]);
    //printf("\n");

    // initialize array
#ifdef USE_FFTW
    for(k=0; k<nfft; k++){ cc[k][0]=0.0; cc[k][1]=0.0;}
#else
    for(k=0; k<2*nfft; k++) cdata[k]=0.0;
#endif

    for(k=0; k<ndata; k++){

        /* set data */
 //               if(st.adbit==2){
 //                   w=re_conv2bit[vd.ch[m].idata[k]];
 //               } else {
 //                   w=vd.ch[m].idata[k];
 //               }
                
#ifdef USE_FFTW
       cc[k][0]=(double)idata[k]; 
       cc[k][1]=sign*(double)qdata[k];
       sumi+=cc[k][0];
       sumq+=cc[k][1];
#else
       cdata[2*k]=(double)idata[k];  
       cdata[2*k+1]=sign*(double)qdata[k];  
       sumi+=cdata[2*k];
       sumq+=cdata[2*k+1];
#endif
       kaisu++;
    }
             

    dci=sumi/(double)kaisu;    // DC component
    dcq=sumq/(double)kaisu;    // DC component
     
    // DC offset removal from data
    for(k=0; k<ndata; k++){
#ifdef USE_FFTW
        cc[k][0]-=dci;
        cc[k][1]-=dcq;
#else
        cdata[2*k]-=dci;
        cdata[2*k+1]-=dcq;
#endif
    }

    // do hanning filtering for spectrum display
    for(k=0; k<nfft; k++){
#ifdef USE_FFTW
        cc[k][0]*=f_hanning[k];  // hanning filter
        cc[k][1]*=f_hanning[k];  // hanning filter
#else
        cdata[2*k]*=f_hanning[k];
        cdata[2*k+1]*=f_hanning[k];
#endif
    }

    // do forward FFT
#ifdef USE_FFTW
    fftw_execute(pfcc);
    for(k=0; k<nfft; k++){
         cc[k][0]*=nfact; cc[k][1]*=nfact;
    }
#else
    cdft(2*nfft,-1,cdata);
#endif

    // do adjustment according to actual sample number

    for(k=0; k<nfft; k++){
#ifdef USE_FFTW
        cspe[2*k]=cc[k][0]*fact;
        cspe[2*k+1]=cc[k][1]*fact;
//        pows=((cc[k][0]*cc[k][0]+cc[k][1]*cc[k][1])*fact); 
#else
        cspe[2*k]=cdata[2*k]*fact;
        cspe[2*k+1]=cdata[2*k+1]*fact;

#endif
        pows=cspe[2*k]*cspe[2*k]+cspe[2*k+1]*cspe[2*k+1];
    }
}



void calc_normal_spectrum(int *idata, int ndata, int nfft, double *f_hanning, double *cspe, 
#ifdef USE_FFTW
                   fftw_complex *cc, fftw_plan pfcc, fftw_plan pbcc)
#else
                   double *cdata)
#endif
{
 /*
    compute normal spectrum 

           2018-02-11
 */



    int k;
    int kaisu;
    double sumi,sumq,dci,dcq;
    double nfact,fact,pows;

    nfact=1.0/(double)nfft;
    fact=(double)nfft/(double)ndata;   // check!!  use of ndata instead of usampl is OK?
    //fact*=fact;

    kaisu=0;
    sumi=0.0;
    sumq=0.0;

    // debug
    //printf("I-DATA");
    //for(k=0; k<100; k++) printf(" %d",idata[k]);
    //printf("\n");
    //printf("Q-DATA");
    //for(k=0; k<100; k++) printf(" %d",qdata[k]);
    //printf("\n");

    // initialize array
#ifdef USE_FFTW
    for(k=0; k<nfft; k++){ cc[k][0]=0.0; cc[k][1]=0.0;}
#else
    for(k=0; k<2*nfft; k++) cdata[k]=0.0;
#endif

    for(k=0; k<ndata; k++){

        /* set data */
 //               if(st.adbit==2){
 //                   w=re_conv2bit[vd.ch[m].idata[k]];
 //               } else {
 //                   w=vd.ch[m].idata[k];
 //               }
                
#ifdef USE_FFTW
       cc[k][0]=(double)idata[k];  
       sumi+=cc[k][0];
#else
       cdata[2*k]=(double)idata[k];  
       sumi+=cdata[2*k];
#endif
       kaisu++;
    }
             

    dci=sumi/(double)kaisu;    // DC component
     
    // DC offset removal from data
    for(k=0; k<ndata; k++){
#ifdef USE_FFTW
        cc[k][0]-=dci;
#else
        cdata[2*k]-=dci;
#endif
    }

    // do hanning filtering for spectrum display
    for(k=0; k<nfft; k++){
#ifdef USE_FFTW
        cc[k][0]*=f_hanning[k];  // hanning filter
        cc[k][1]*=f_hanning[k];  // hanning filter
#else
        cdata[2*k]*=f_hanning[k];
        cdata[2*k+1]*=f_hanning[k];
#endif
    }

    // do forward FFT
#ifdef USE_FFTW
    fftw_execute(pfcc);
    for(k=0; k<nfft; k++){
         cc[k][0]*=nfact; cc[k][1]*=nfact;
    }
#else
    cdft(2*nfft,-1,cdata);
#endif

    // do adjustment according to actual sample number

    for(k=0; k<nfft; k++){
#ifdef USE_FFTW
        cspe[2*k]=cc[k][0]*fact;
        cspe[2*k+1]=cc[k][1]*fact;
        //pows=((cc[k][0]*cc[k][0]+cc[k][1]*cc[k][1])*fact); 
#else
        cspe[2*k]=cdata[2*k]*fact;
        cspe[2*k+1]=cdata[2*k+1]*fact;
#endif
        pows=cspe[2*k]*cspe[2*k]+cspe[2*k+1]*cspe[2*k+1];
    }
}



void iq_to_real(int flag_lsb, int *idata, int *qdata, int ndata, int *rdata)
{
 /*
    compute combined spectrum of LSB and USB from I and Q time siries data 
    then IFFT to obtaine real data in double sampling speed

    size of radata is 2*ndata
 */



    int n,k;
    int kaisu;
    double sumi,sumq,dci,dcq;
    double nfact,fact,pows;
    int nloop;
    int nfft,nfft2;
    double cdata[2048];
    double  cspe[4096];
    double wrdata[4096];
    double sum,sum2,ave,sigma;
    int flag_usb;
    double amax,amin;

    //nfft=1024;
    nfft=128;
    nfft2=2*nfft;

    amin=0.0; amax=0.0;
    nfact=1.0/(double)nfft;
    fact=(double)nfft/(double)ndata;   // check!!  use of ndata instead of usampl is OK?
    //fact*=fact;

    nloop=ndata/nfft + 1;

    //flag_usb=1;   // for USB
    flag_usb=0;   // for LSB

    for(n=0; n< nloop; n++){
        for(k=0; k<2*nfft; k++) cdata[k]=0.0;
        for(k=0; k<4*nfft; k++) cspe[k]=0.0;
        kaisu=0;
        sumi=0.0;
        sumq=0.0;
        for(k=0; k<nfft; k++){
            if(n*nfft+k >= ndata) break;
            cdata[2*k]  =idata[n*nfft+k];
            cdata[2*k+1]=qdata[n*nfft+k];
            sumi+=cdata[2*k];
            sumq+=cdata[2*k+1];
            kaisu++;
        }
        dci=sumi/(double)kaisu;    // DC component
        dcq=sumq/(double)kaisu;    // DC component
     
        // DC offset removal from data
        for(k=0; k<kaisu; k++){
          cdata[2*k]-=dci;
          cdata[2*k+1]-=dcq;
        }
        cdft(2*nfft,-1,cdata);

        // do adjustment according to actual sample number


        // USB
        if(flag_lsb==0){     // USB
            for(k=1; k<nfft; k++){
               //cspe[2*k]=cdata[2*k]*fact;
               //cspe[2*k+1]=cdata[2*k+1]*fact;
               cspe[2*k]=cdata[2*k];
               cspe[2*k+1]=cdata[2*k+1];
            }
        } else {            // LSB
            // LSB
            for(k=1; k<nfft; k++){
               //cspe[2*k]=cdata[2*k]*fact;
               //cspe[2*k+1]=cdata[2*k+1]*fact;
               cspe[2*k]=cdata[2*nfft-2*k];
               cspe[2*k+1]=cdata[2*nfft-2*k+1];
            }
        }

       
       // conjugate copy
        for(k=1; k<nfft-1; k++){
           cspe[2*nfft2-2*k]=cspe[2*k];
           cspe[2*nfft2-2*k+1]=-cspe[2*k+1];
           //cspe[2*nfft2-2*k]=cdata[2*k];       // this is NG why??
           //cspe[2*nfft2-2*k+1]=-cdata[2*k+1];
        }

        cspe[0]=0.0;              // band edge 0
        cspe[1]=0.0;              //
        cspe[2*(nfft-1)]=0.0;    // band edge 0
        cspe[2*(nfft-1)+1]=0.0;  //

        // IFFT
        cdft(2*nfft2,1,cspe);

        // convert to real time siries data;
        sum=0.0; sum2=0.0;
        for(k=0; k<2*kaisu; k++){
            wrdata[k]=cspe[2*k]*nfft;    // real data
            sum+=wrdata[k];
            sum2+=wrdata[k]*wrdata[k];
        }
        ave=sum/(double)(2*kaisu);
        sigma=sqrt((sum2-ave*ave)/(double)(2*kaisu));
        for(k=0; k<2*kaisu; k++){
            if(wrdata[k] > ave+sigma){
               rdata[n*2*nfft+k]=3;
            } else if(wrdata[k] > ave && wrdata[k] <= ave+sigma){
                rdata[n*2*nfft+k]=2;
            } else if(wrdata[k] > ave-sigma && wrdata[k] <= ave){
                rdata[n*2*nfft+k]=1;
            } else {
                rdata[n*2*nfft+k]=0;
            }
        }

    }

}



void decode_IQ(int flag_lsb, int *idata, int *qdata, int ndata, int *sphase, int *rdata)
{
/*

   generate real data from I/Q data

   # of rdata is 2*ndata
   sphase -- start phase should be 0,1,2,3




*/
    int n;
    int scount;
    int phase;
    int ipq,imq,iq;
    int sign;

    int ic,qs;
    //int is,qc;
    //int icos[4]={1,0,-1,0};
    //int isin[4]={0,1,0,-1};
    int icos[4]={1,-1,1,-1};
    int isin[4]={1,-1,1,-1};

    phase=(*sphase % 4);

    sign=1;                    // USB
    if(flag_lsb==1) sign=-1;   // LSB

    for(n=0; n<2*ndata; n++){
        phase=(phase % 4);
        // test quadrature mixer
        ic=(2*idata[n/2]-3)*icos[phase];   // test quadrature mixer
        qs=(2*qdata[n/2]-3)*isin[phase];  // test quadrature mixer

        // tset nothing to do 
        //ic=idata[n/2];                      
        //qs=qdata[n/2];

        rdata[n]=ic+sign*qs;

//        is=(2*idata[n/2]-3)*isin[phase]; 
//        qs=(2*qdata[n/2]-3)*icos[phase];

//        rdata[n]=imq*icos[phase]-(qs+is)*isin[phase];

        // test sqrt(I*I+Q*Q)
        //rdata[n]=(int)sqrt(((double)idata[n/2]-1.5)*((double)idata[n/2]-1.5)+((double)qdata[n/2]-1.5)*((double)qdata[n/2]-1.5));
        phase++;
    }
    *sphase=(phase % 4);
}



void decode_IQ_HF(int flag_lsb, int *idata, int *qdata, int ndata, int *rdata)
{
/*

   generate real data from I/Q data using Hilbert Filter

   2018-03-08

   filter coefficient is got from Takeuchi, 2010evlbi

   This is valid only for 2bit AD case
    

   [in]
     idata[ndata]  ---  I data   
     qdata[ndata]  ---  Q data
     ndata         ---  size of I/Q data

   [out]
     rdata[2*ndata] --  real data
                        size is 2*ndata



*/
    int n,k;
    //int sum;
    double sum;
    int idat;
    double sumi,sumi2,sumq,sumq2;
    double avei,sigmai,aveq,sigmaq;
    //int flag_lsb=1;
    double fact=47.1239;      // 15*PI

    //double hc[11]={-0.4,0.0, -0.6666667,0.0, -2.0, 0.0, 
    //                2.0, 0.0,0.6666667, 0.0,0.4};   // Hilber filter coeff * PI 
    //int hc[11]={-6,0,-10,0,-30,0,30,0,10,0,6};   // Hilber filter coeff * PI * 15  phase shift by -90 degree
                                                 // because time axis reversed for VLBI data
    //double hc[11]={-0.12732395,0,-0.21220659,0,-0.63661977,
    //                0,0.63661977,0,0.21220659,0,0.12732395};   // Hilber filter coeff  phase shift by -90 degree
    //                 // -2/5PI,0,-2/3PI,0,-2/PI,0,2/PI,0,2/3PI,0,2/5PI
    double hf[3]={0.63661977,0.21220659,0.12732395};   // {2/PI,2/3PI,2/5PI}

    sumi=0.0; sumi2=0.0; sumq=0; sumq2=0.0;
    for(n=0; n<2*ndata; n++){
        
        sum=0;    // sum is qdat!!
        if(n >= 5 && n < 2*ndata-6){
            if( (n % 2) == 1){
               //for(k=0; k<11; k++){
               // //sum+=qdat2[n-5+k]*hc[10-k];
               //    sum+=(2*qdata[(n-5+k)/2]-3)*hc[10-k];
               //}
               //sum/=fact;

               //sum=hf[0]*((2*qdata[(n-1)/2]-3)-(2*qdata[(n+1)/2]-3))
               //    +hf[1]*((2*qdata[(n-3)/2]-3)-(2*qdata[(n+3)/2]-3))
               //    +hf[2]*((2*qdata[(n-5)/2]-3)-(2*qdata[(n+5)/2]-3));
               
               sum=hf[0]*((2*qdata[(n+1)/2]-3)-(2*qdata[(n-1)/2]-3))
                   +hf[1]*((2*qdata[(n+3)/2]-3)-(2*qdata[(n-3)/2]-3))
                   +hf[2]*((2*qdata[(n+5)/2]-3)-(2*qdata[(n-5)/2]-3));
            }
        }

        idat=2*idata[n/2]-3;
        if( (n % 2) == 1) idat=0;

        sumq+=sum;
        sumq2+=sum*sum;
        sumi+=(2*idata[n/2]-3);
        sumi2+=(2*idata[n/2]-3)*(2*idata[n/2]-3);

        if(flag_lsb==1){
           rdata[n]=idat-sum;         //  advanced by 90 deg
        } else {
           rdata[n]=idat+sum;         //  delayed by 90 degree
        }
    }

    avei=sumi/(double)(2*ndata);
    aveq=sumq/(double)(2*ndata);
    sigmai=sqrt(sumi2/(double)(2*ndata) - avei*avei);
    sigmaq=sqrt(sumq2/(double)(2*ndata) - aveq*aveq);

    // this to find best coeff 
    //printf(" avei, sigmai, aveq, sigmaq,  sigmaq/sigmai:  %f  %f  %f  %f  %f\n",
    //    avei,sigmai,aveq,sigmaq, sigmaq/sigmai);
    //printf(" sigmaq/sigmai:  %f\n",sigmaq/sigmai);



}


void decode_IQ_HF_bup(int *idata, int *qdata, int ndata, int *rdata)
{
/*

   generate real data from I/Q data using Hilbert Filter

   2018-03-08

   filter coefficient is got from Takeuchi, 2010evlbi

   This is valid only for 2bit AD case
    

   [in]
     idata[ndata]  ---  I data   
     qdata[ndata]  ---  Q data
     ndata         ---  size of I/Q data

   [out]
     rdata[2*ndata] --  real data
                        size is 2*ndata



*/
    int n,k;
    int sum;
    double sumi,sumi2,sumq,sumq2;
    double avei,sigmai,aveq,sigmaq;
    int flag_lsb=1;

    //double hc[11]={-0.4,0.0, -0.6666667,0.0, -2.0, 0.0, 
    //                2.0, 0.0,0.6666667, 0.0,0.4};   // Hilber filter coeff * PI 
    int hc[11]={-6,0,-10,0,-30,0,30,0,10,0,6};   // Hilber filter coeff * PI * 15 

    sumi=0.0; sumi2=0.0; sumq=0; sumq2=0.0;
    for(n=0; n<2*ndata; n++){
        
        sum=0;
        if(n >= 5 && n < 2*ndata-6){
           for(k=0; k<11; k++){
            //sum+=qdat2[n-5+k]*hc[k];
               sum+=(2*qdata[(n-5+k)/2]-3)*hc[k];
           }
           sum/=40;         // coeff 40 is empirical
        }
        sumq+=sum;
        sumq2+=sum*sum;
        sumi+=(2*idata[n/2]-3);
        sumi2+=(2*idata[n/2]-3)*(2*idata[n/2]-3);

        if(flag_lsb==1){
           rdata[n]=(2*idata[n/2]-3)-sum;         // negative sign is essential/
        } else {
           rdata[n]=(2*idata[n/2]-3)+sum; 
        }
    }

    avei=sumi/(double)(2*ndata);
    aveq=sumq/(double)(2*ndata);
    sigmai=sqrt(sumi2/(double)(2*ndata) - avei*avei);
    sigmaq=sqrt(sumq2/(double)(2*ndata) - aveq*aveq);

    // this to find best coeff 
    //printf(" avei, sigmai, aveq, sigmaq,  sigmaq/sigmai:  %f  %f  %f  %f  %f\n",
    //    avei,sigmai,aveq,sigmaq, sigmaq/sigmai);
    //printf(" sigmaq/sigmai:  %f\n",sigmaq/sigmai);



}


void decode_IQ_HF_old(int *idata, int *qdata, int ndata, int *rdata)
{
/*

   generate real data from I/Q data using Hilbert Filter

   # of rdata is 2*ndata




*/
    int n,k;
    int scount;
    int phase;
    int ipq,imq,iq;
    int *idat2,*qdat2;
    int *qdath,*wdat;
    double sum,sumi,sumq;
    double wd,wmax,wmin;

    double ic,qs;
    //double hc[11]={-0.4,0.0, -0.6666667,0.0, -2.0, 0.0, 
    //                2.0, 0.0,0.6666667, 0.0,0.4};   // Hilber filter coeff * PI 
    double hc[11]={-6,0,-10,0,-30,0,30,0,10,0,6};   // Hilber filter coeff * PI * 15 

    idat2=(int *)malloc(2*ndata*sizeof(int));
    qdat2=(int *)malloc(2*ndata*sizeof(int));
    qdath=(int *)malloc(2*ndata*sizeof(int));
    wdat =(int *)malloc(2*ndata*sizeof(int));
    sumi=0.0;
    sumq=0.0;

    for(n=0; n<2*ndata; n++){
        idat2[n]=(2*idata[n/2]-3);
        sumi+=idat2[n];
        qdat2[n]=(2*qdata[n/2]-3);
        sumq+=qdat2[n];
        qdath[n]=0.0;
    }
    sumi/=(double)(2*ndata);
    sumq/=(double)(2*ndata);
    //for(n=0; n<2*ndata; n++){
    //    idat2[n]-=sumi;
    //    qdat2[n]-=sumq;
    //}

    wmin=0.0; wmax=0.0;
    for(n=5; n<2*ndata-6; n++){
        sum=0.0;
        for(k=0; k<11; k++){
            sum+=qdat2[n-5+k]*hc[k];
        }
        //qdath[n]=sum*PI;
        qdath[n]=sum;
        if(qdath[n] < wmin) wmin=qdath[n];
        if(qdath[n] > wmax) wmax=qdath[n];
    }


    for(n=0; n<2*ndata; n++){
        wdat[n]=idat2[n]+qdath[n]/50;
        //wdat[n]=idat2[n]-qdath[n]/92.0;
        rdata[n]=wdat[n];
    }

    free(idat2);
    free(qdat2);
    free(qdath);
    free(wdat);
}




void save_IQ_for_debug(int *idata, int *qdata, int ndata, int mode)
{
/*

   save IQ data for debug

      2018-04-30

      mode --  0  open as a new file 
               +VE  open as an append mode

*/
    int n;
    int ierr;
    char *filename="IQ_save.txt";
    FILE *fo;

    if(mode == 0){
       if( (fo  = fopen(filename, "w" )) == NULL ){
           printf("IQ save file (%s) Open Error! \n",filename);
           ierr=-1;
           goto exit2return;
           printf("IQ save file (%s)  Opened! \n",filename);
       }
    } else {
       if( (fo  = fopen(filename, "a" )) == NULL ){
           printf("IQ save file (%s) Open Error! \n",filename);
           ierr=-1;
           goto exit2return;
       }
    }
    fprintf(fo,"%d\n",ndata);

    for(n=0; n<ndata; n++){
        fprintf(fo,"%d  %d  %d\n",n+1,idata[n],qdata[n]);
    }
    fclose(fo);
exit2return:;
}

